#
#   Copyright 2026 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from hopsworks_common import client
from hsfs.core.chart import Chart


class ChartApi:
    def create_chart(self, chart: Chart) -> None:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "charts",
        ]

        _client._send_request(
            "POST",
            path_params,
            headers={"content-type": "application/json"},
            data=chart.json(),
        )

    def get_charts(self) -> list[Chart]:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "charts",
        ]

        return Chart.from_response_json(
            _client._send_request(
                "GET",
                path_params,
                headers={"content-type": "application/json"},
            )
        )

    def get_chart(self, chart_id: int) -> Chart:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "charts",
            chart_id,
        ]

        return Chart.from_response_json(
            _client._send_request(
                "GET",
                path_params,
                headers={"content-type": "application/json"},
            )
        )

    def update_chart(self, chart: Chart) -> None:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "charts",
            chart._id,
        ]

        _client._send_request(
            "PUT",
            path_params,
            headers={"content-type": "application/json"},
            data=chart.json(),
        )

    def delete_chart(self, chart_id: int) -> None:
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "charts",
            chart_id,
        ]

        _client._send_request(
            "DELETE",
            path_params,
            headers={"content-type": "application/json"},
        )



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
import subprocess
import tempfile
from pathlib import Path
from queue import Empty, Queue
from threading import Thread
from typing import TYPE_CHECKING

from hopsworks.mcp.utils.tags import TAGS


if TYPE_CHECKING:
    from fastmcp import FastMCP


def enqueue_output(out, queue):
    # TODO: obtain output byte-by-byte
    for line in iter(out.readline, b""):
        queue.put(line)
    out.close()


class TerminalTools:
    """Tools implementing shell and terminal access to Hopsworks."""

    def __init__(self, mcp: FastMCP):
        """Initialize the TerminalTools with the MCP server instance.

        Parameters:
            mcp: The MCP server instance
        """
        self.mcp = mcp
        self.mcp.tool(tags=[TAGS.TERMINAL])(self.start_session)
        self.mcp.tool(tags=[TAGS.TERMINAL])(self.add_input)
        self.mcp.tool(tags=[TAGS.TERMINAL])(self.get_output)

        self.sessions: dict[int, tuple[subprocess.Popen, Queue, str, str]] = {}

    def start_session(self, cwd: str) -> int:
        """Start a terminal session in the specified directory.

        Parameters:
            cwd: The directory path to start the session in.

        Returns:
            The session process identifier (pid).
        """
        # TODO: delete processes which were used longer than 5 min ago

        envdir = tempfile.mkdtemp()

        proc = subprocess.Popen(
            ["bash"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=cwd,
            text=True,
            bufsize=1,
            env={
                "PROMPT_COMMAND": f'python -c "import json, os; print(json.dumps(dict(os.environ)))" > {envdir}/env.json'
            },
        )

        output_queue = Queue()
        t = Thread(target=enqueue_output, args=(proc.stdout, output_queue))
        t.daemon = True  # thread dies with the program
        t.start()

        self.sessions[proc.pid] = (proc, output_queue, "", envdir)
        # TODO: check for errors

        return proc.pid

    def add_input(self, pid: int, addon: str):
        r"""Add input to the terminal session.

        Note that if you want to simulate pressing Enter, you need to include a newline character.
        So, to run a command, you should set addon to something like "ls -la\n".

        Parameters:
            pid: The process identifier of the terminal session.
            addon: The input string to add to the session.
        """
        proc, _, _, _ = self.sessions[pid]
        if not proc.stdin:
            raise Exception("Process stdin is not available")
        proc.stdin.write(addon)
        proc.stdin.flush()

    def get_output(self, pid: int, offset: int = 0) -> str:
        r"""Get output from the terminal session.

        Set offset to the number of characters you have already retrieved, so that only new output is returned.

        Parameters:
            pid: The process identifier of the terminal session.
            offset: The offset from which to retrieve new output.

        Returns:
            The output string from the session starting from the specified offset.
        """
        proc, output_queue, output, envdir = self.sessions[pid]
        while True:
            try:
                output += output_queue.get_nowait()
            except Empty:
                self.sessions[pid] = (proc, output_queue, output, envdir)
                return output[offset:]

    def get_environ(self, pid: int) -> dict[str, str]:
        r"""Get the current values of the environment variables of the terminal session.

        For example, this way you can figure out what the current working directory is by checking the "PWD" variable.

        Parameters:
            pid: The process identifier of the terminal session.

        Returns:
            A dictionary of environment variables.
        """
        _, _, _, envdir = self.sessions[pid]
        return json.loads((Path(envdir) / "env.json").read_text())



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""MCP server for Hopsworks."""

from fastmcp import FastMCP
from starlette import status
from starlette.responses import Response

from .prompts import ProjectPrompts, SystemPrompts
from .resources.project import ProjectResources
from .tools import (
    AuthTools,
    BrewerTools,
    DatasetTools,
    FeatureGroupTools,
    JobTools,
    ProjectTools,
    TerminalTools,
)


# Create a FastMCP server instance
mcp = FastMCP(name="Hopsworks MCP")

# Initialize tools and resources
AuthTools(mcp)
ProjectTools(mcp)
ProjectResources(mcp)
ProjectPrompts(mcp)
SystemPrompts(mcp)
JobTools(mcp)
DatasetTools(mcp)
FeatureGroupTools(mcp)
TerminalTools(mcp)
BrewerTools(mcp)


@mcp.custom_route("/health", methods=["GET"])
async def health(_):
    return Response(status_code=status.HTTP_204_NO_CONTENT)


app = mcp.http_app()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from hopsworks_apigen import public
from hopsworks_common import usage
from hopsworks_common.core.constants import (
    HAS_TRINO,
    trino_not_installed_message,
)
from hopsworks_common.core.trino_api import (
    DEFAULT_SOURCE,
    DEFAULT_SQLALCHEMY_SOURCE,
    TrinoApi,
)


if HAS_TRINO:
    from trino import constants
    from trino.transaction import IsolationLevel
else:
    raise ImportError(trino_not_installed_message)

if TYPE_CHECKING:
    from sqlalchemy.engine import Engine
    from trino.dbapi import Connection


@public("hopsworks.core.trino_dbapi.connect")
@usage.method_logger
def connect(
    source: str = DEFAULT_SOURCE,
    catalog: str = constants.DEFAULT_CATALOG,
    schema: str = constants.DEFAULT_SCHEMA,
    session_properties: dict | None = None,
    http_headers: dict | None = None,
    max_attempts: int = constants.DEFAULT_MAX_ATTEMPTS,
    request_timeout: int = constants.DEFAULT_REQUEST_TIMEOUT,
    isolation_level: IsolationLevel = IsolationLevel.AUTOCOMMIT,
    verify: bool | str = True,
    http_session: Any = None,
    client_tags: list[str] | None = None,
    legacy_primitive_types: bool = False,
    legacy_prepared_statements: bool | None = None,
    roles: dict | None = None,
    timezone: str | None = None,
    encoding: str | list[str] | None = None,
) -> Connection:
    """Connect to Trino using the native DBAPI interface.

    Use this when you want to work with cursors and the native Python DB API.
    For SQLAlchemy integration, use `create_engine()` instead.

    Hopsworks automatically handles authentication using project credentials from the secrets storage.
    The connection is configured to use HTTPS with self-signed certificates.

    Parameters:
        source: Source identifier for Trino queries.
        catalog: Trino catalog to connect to.
        schema: Database schema within the catalog.
        session_properties: Dictionary of Trino session properties.
        http_headers: Additional HTTP headers for the connection.
        max_attempts: Maximum number of retry attempts for failed requests.
        request_timeout: Timeout in seconds for each HTTP request.
        isolation_level: Transaction isolation level.
        verify: Whether to verify SSL certificates.
                Set verify="/path/to/cert.crt" if you want to verify the ssl cert (default: True).
        http_session: Custom HTTP session for connection pooling.
        client_tags: Tags to identify the client in Trino query logs.
        legacy_primitive_types: Whether to use legacy primitive type handling.
        legacy_prepared_statements: Whether to use legacy prepared statement handling.
        roles: Dictionary mapping catalog names to role names.
        timezone: Timezone for the session.
        encoding: Character encoding for the connection.

    Returns:
        A connection object implementing the Python DB API 2.0 specification.

    Raises:
        hopsworks_common.client.exceptions.TrinoException: If the service discovery domain is not configured.
        hopsworks.client.exceptions.RestAPIError: If credentials cannot be retrieved.

    Example:
        ```python
        import hopsworks
        from hopsworks.core.trino_dbapi import connect

        project = hopsworks.login()
        conn = connect(catalog="iceberg", schema="my_db")
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM my_table")
        rows = cursor.fetchall()
        for row in rows:
            print(row)
        ```
    """
    trino_api = TrinoApi()
    return trino_api.connect(
        source=source,
        catalog=catalog,
        schema=schema,
        session_properties=session_properties,
        http_headers=http_headers,
        max_attempts=max_attempts,
        request_timeout=request_timeout,
        isolation_level=isolation_level,
        verify=verify,
        http_session=http_session,
        client_tags=client_tags,
        legacy_primitive_types=legacy_primitive_types,
        legacy_prepared_statements=legacy_prepared_statements,
        roles=roles,
        timezone=timezone,
        encoding=encoding,
    )


@public("hopsworks.core.trino_dbapi.create_engine")
@usage.method_logger
def create_engine(
    source: str = DEFAULT_SQLALCHEMY_SOURCE,
    catalog: str = constants.DEFAULT_CATALOG,
    schema: str = constants.DEFAULT_SCHEMA,
    session_properties: dict | None = None,
    http_headers: dict | None = None,
    max_attempts: int = constants.DEFAULT_MAX_ATTEMPTS,
    request_timeout: int = constants.DEFAULT_REQUEST_TIMEOUT,
    isolation_level: IsolationLevel = IsolationLevel.AUTOCOMMIT,
    verify: bool | str = True,
    http_session: Any = None,
    client_tags: list[str] | None = None,
    legacy_primitive_types: bool = False,
    legacy_prepared_statements: bool | None = None,
    roles: dict | None = None,
    timezone: str | None = None,
    encoding: str | list[str] | None = None,
) -> Engine:
    """Create a SQLAlchemy engine for Trino.

    Use this when you want to work with SQLAlchemy for database operations.
    For the native Python DB API, use `connect()` instead.

    Hopsworks automatically handles authentication using project credentials from the secrets storage.
    The connection is configured to use HTTPS with self-signed certificates.

    Parameters:
        source: Source identifier for Trino queries.
        catalog: Trino catalog to connect to.
        schema: Database schema within the catalog.
        session_properties: Dictionary of Trino session properties.
        http_headers: Additional HTTP headers for the connection.
        max_attempts: Maximum number of retry attempts for failed requests.
        request_timeout: Timeout in seconds for each HTTP request.
        isolation_level: Transaction isolation level.
        verify: Whether to verify SSL certificates.
                Set verify="/path/to/cert.crt" if you want to verify the ssl cert (default: True).
        http_session: Custom HTTP session for connection pooling.
        client_tags: Tags to identify the client in Trino query logs.
        legacy_primitive_types: Whether to use legacy primitive type handling.
        legacy_prepared_statements: Whether to use legacy prepared statement handling.
        roles: Dictionary mapping catalog names to role names.
        timezone: Timezone for the session.
        encoding: Character encoding for the connection.

    Returns:
        An Engine object implementing the SQLAlchemy interface.

    Raises:
        hopsworks_common.client.exceptions.TrinoException: If the service discovery domain is not configured.
        hopsworks.client.exceptions.RestAPIError: If credentials cannot be retrieved.

    Example:
        ```python
        import hopsworks
        from hopsworks.core.trino_dbapi import create_engine
        from sqlalchemy import text

        project = hopsworks.login()
        engine = create_engine(catalog="iceberg", schema="my_db")
        with engine.connect() as connection:
            cursor = connection.execute(text("SELECT * FROM tiny.nation")).cursor
            rows = cursor.fetchall()
            for row in rows:
                print(row)
        ```
    """
    trino_api = TrinoApi()
    return trino_api.create_engine(
        source=source,
        catalog=catalog,
        schema=schema,
        session_properties=session_properties,
        http_headers=http_headers,
        max_attempts=max_attempts,
        request_timeout=request_timeout,
        isolation_level=isolation_level,
        http_session=http_session,
        client_tags=client_tags,
        legacy_primitive_types=legacy_primitive_types,
        legacy_prepared_statements=legacy_prepared_statements,
        verify=verify,
        roles=roles,
        timezone=timezone,
        encoding=encoding,
    )



#
#   Copyright 2023 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar


if TYPE_CHECKING:
    from hopsworks_common.core.constants import HAS_NUMPY

    if HAS_NUMPY:
        import numpy as np
    import pandas as pd
    from hsfs.core.job import Job
    from hsfs.validation_report import ValidationReport


class FeatureGroupWriter:
    def __init__(self, feature_group):
        self._feature_group = feature_group

    def __enter__(self):
        return self

    def insert(
        self,
        features: pd.DataFrame
        | TypeVar("pyspark.sql.DataFrame")
        | TypeVar("pyspark.RDD")
        | np.ndarray
        | list[list],
        overwrite: bool | None = False,
        operation: str | None = "upsert",
        storage: str | None = None,
        write_options: dict[str, Any] | None = None,
        validation_options: dict[str, Any] | None = None,
        transformation_context: dict[str, Any] = None,
        transform: bool = True,
    ) -> tuple[Job | None, ValidationReport | None]:
        if validation_options is None:
            validation_options = {}
        if write_options is None:
            write_options = {}
        return self._feature_group.insert(
            features=features,
            overwrite=overwrite,
            operation=operation,
            storage=storage,
            write_options={"start_offline_materialization": False, **write_options},
            validation_options={"fetch_expectation_suite": False, **validation_options},
            transformation_context=transformation_context,
            transform=transform,
        )

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._feature_group.finalize_multi_part_insert()



#
#   Copyright 2026 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
"""Trino API for connecting to Trino from within Hopsworks.

This module provides convenience functions for connecting to Trino
using either the native DBAPI interface or SQLAlchemy.
Authentication and connection configuration are handled automatically.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from hopsworks_apigen import public
from hopsworks_common import client, usage
from hopsworks_common.client.exceptions import TrinoException
from hopsworks_common.core import project_api, secret_api
from hopsworks_common.core.constants import HAS_TRINO
from hopsworks_common.core.variable_api import VariableApi
from hopsworks_common.decorators import uses_trino


if HAS_TRINO:
    from trino.auth import BasicAuthentication
    from trino.constants import (
        DEFAULT_CATALOG,
        DEFAULT_MAX_ATTEMPTS,
        DEFAULT_REQUEST_TIMEOUT,
        DEFAULT_SCHEMA,
    )
    from trino.dbapi import connect as _trino_connect
    from trino.sqlalchemy import URL
    from trino.transaction import IsolationLevel

    AUTOCOMMIT = IsolationLevel.AUTOCOMMIT
else:
    DEFAULT_CATALOG = None
    DEFAULT_SCHEMA = None
    DEFAULT_MAX_ATTEMPTS = None
    DEFAULT_REQUEST_TIMEOUT = None
    AUTOCOMMIT = None


if TYPE_CHECKING:
    from hopsworks_common import project
    from sqlalchemy.engine import Engine
    from trino.dbapi import Connection
    from trino.transaction import IsolationLevel  # noqa: TC004


_logger = logging.getLogger(__name__)

TRINO_SERVICE_NAME = "coordinator.trino.service"
TRINO_PORT = 8443
DEFAULT_SOURCE = "hopsworks-trino-python-client"
DEFAULT_SQLALCHEMY_SOURCE = "hopsworks-trino-sqlalchemy"
HTTPS = "https"


@public("hopsworks.core.trino_api.TrinoAPI")
class TrinoApi:
    """API for connecting to Trino from within Hopsworks.

    This class provides methods to establish connections to Trino using either
    the native Trino DBAPI or SQLAlchemy engine.
    Authentication is handled automatically using Hopsworks project credentials
    stored in the secrets storage.

    The connection configuration adapts based on whether you're connecting from
    within the Hopsworks cluster or externally through the load balancer.

    Example:
        ```python
        import hopsworks
        from sqlalchemy import text

        project = hopsworks.login()
        trino_api = project.get_trino_api()  # Get an instance of TrinoApi from the project context
        conn = trino_api.connect(catalog="iceberg", schema="my_db")  # Get a DBAPI connection
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM my_table")
        rows = cursor.fetchall()
        for row in rows:
            print(row)

        # Or using SQLAlchemy
        engine = trino_api.create_engine(catalog="iceberg", schema="my_db")
        with engine.connect() as connection:
            result = connection.execute(text("SELECT * FROM my_table"))
            for row in result:
                print(row)
        ```
    """

    def __init__(self, project: project.Project | None = None):
        """Initialize the TrinoApi instance.

        Parameters:
            project: The Hopsworks project context. If None, the default project is used.

        """
        self._variable_api: VariableApi = VariableApi()
        self._service_discovery_domain = (
            self._variable_api.get_service_discovery_domain()
        )
        self._secret_api: secret_api.SecretsApi = secret_api.SecretsApi()
        self._project_api: project_api.ProjectApi = project_api.ProjectApi()
        if project is None:
            _client = client.get_instance()
            self.project_name = _client._project_name
        else:
            self.project_name = project.name

    def _get_ca_chain_path(self, verify: bool | str) -> bool | str:
        """Get the SSL certificate or pass through a custom verify value.

        Parameters:
            verify: Whether to verify the SSL certificate, or a path to a custom CA bundle.

        Returns:
            The file path of the downloaded SSL certificate if it was downloaded.
            The original verify value if verification is handled by the caller or disabled.
        """
        if not client._is_external() and verify is True:
            _client = client.get_instance()
            return _client._get_ca_chain_path()
        return verify

    @public
    @usage.method_logger
    def get_host(self) -> str:
        """Retrieve the Trino host based on client location.

        Returns the external load balancer domain if connecting from outside
        the cluster, otherwise returns the internal service discovery hostname.

        Returns:
            The Trino host URL.

        Raises:
            hopsworks_common.client.exceptions.TrinoException:
                If service discovery domain is not configured for internal clients.
        """
        _logger.debug("Retrieving Trino host.")
        if client._is_external():
            external_domain = self._variable_api.get_loadbalancer_external_domain(
                "trino"
            )
            host = external_domain
        else:
            if self._service_discovery_domain == "":
                raise TrinoException(
                    "Client could not get Trino hostname from service_discovery_domain. "
                    "The variable is either not set or empty in Hopsworks cluster configuration."
                )
            host = f"{TRINO_SERVICE_NAME}.{self._service_discovery_domain}"
        _logger.debug(f"Connecting to Trino on host {host} and port {TRINO_PORT}.")
        return host

    @public
    @usage.method_logger
    def get_port(self) -> int:
        """Get the Trino port number.

        Returns:
            The port number for connecting to Trino.
        """
        return TRINO_PORT

    def _get_password(self, user: str) -> str:
        """Retrieve the password for the given user from secrets storage.

        Parameters:
            user: The username for which to retrieve the password.

        Returns:
            The password for the given user.

        Raises:
            hopsworks_common.client.exceptions.TrinoException:
                If the password cannot be retrieved.
        """
        secret = self._secret_api.get_secret(user)
        if secret is None:
            raise TrinoException(
                f"Client could not retrieve credentials for user {user} from secrets storage. "
                "Ensure that the secret exists and is accessible."
            )
        return secret.value

    @public
    @usage.method_logger
    def get_basic_auth(self) -> tuple[str, str]:
        """Get a tuple containing the username and password for the current project user.

        Returns:
            A tuple containing the username and the password for the current project user.

        Raises:
            hopsworks_common.client.exceptions.TrinoException:
                If credentials cannot be retrieved from secrets storage or
                if the client cannot determine the username for the current project user.
        """
        username = self._project_api.get_user_info().get("username", None)
        if username is None:
            raise TrinoException(
                "Client could not determine username for the current project user."
            )
        user = f"{self.project_name}__{username}"
        password = self._get_password(user)
        return user, password

    @uses_trino
    @public
    @usage.method_logger
    def connect(
        self,
        source: str = DEFAULT_SOURCE,
        catalog: str = DEFAULT_CATALOG,
        schema: str = DEFAULT_SCHEMA,
        session_properties: dict | None = None,
        http_headers: dict | None = None,
        max_attempts: int = DEFAULT_MAX_ATTEMPTS,
        request_timeout: int = DEFAULT_REQUEST_TIMEOUT,
        isolation_level: IsolationLevel = AUTOCOMMIT,
        verify: bool | str = True,
        http_session: Any = None,
        client_tags: list[str] | None = None,
        legacy_primitive_types: bool = False,
        legacy_prepared_statements: bool | None = None,
        roles: dict | None = None,
        timezone: str | None = None,
        encoding: str | list[str] | None = None,
    ) -> Connection:
        """Connect to Trino using the native DBAPI interface.

        Parameters:
            source: Source identifier for Trino queries.
            catalog: Trino catalog to connect to.
            schema: Database schema within the catalog.
            session_properties: Dictionary of Trino session properties.
            http_headers: Additional HTTP headers for the connection.
            max_attempts: Maximum number of retry attempts for failed requests.
            request_timeout: Timeout in seconds for each HTTP request.
            isolation_level: Transaction isolation level.
            verify: Whether to verify SSL certificates.
                    Set verify="/path/to/cert.crt" if you want to verify the ssl cert (default: True).
            http_session: Custom HTTP session for connection pooling.
            client_tags: Tags to identify the client in Trino query logs.
            legacy_primitive_types: Whether to use legacy primitive type handling.
            legacy_prepared_statements: Whether to use legacy prepared statement handling.
            roles: Dictionary mapping catalog names to role names.
            timezone: Timezone for the session.
            encoding: Character encoding for the connection.

        Returns:
            A connection object implementing the Python DB API 2.0 specification.

        Raises:
            hopsworks_common.client.exceptions.TrinoException:
                If service discovery domain is not configured.
            hopsworks.client.exceptions.RestAPIError:
                If credentials cannot be retrieved.

        Example:
            ```python
            import hopsworks

            project = hopsworks.login()
            trino_api = project.get_trino_api()
            conn = trino_api.connect(catalog="iceberg", schema="my_db")
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM my_table")
            rows = cursor.fetchall()
            for row in rows:
                print(row)
            ```
        """
        host = self.get_host()
        port = self.get_port()

        user, password = self.get_basic_auth()
        basic_auth = BasicAuthentication(user, password)

        return _trino_connect(
            host=host,
            port=port,
            user=user,
            catalog=catalog,
            schema=schema,
            source=source,
            auth=basic_auth,
            http_scheme=HTTPS,
            verify=self._get_ca_chain_path(verify),
            session_properties=session_properties,
            http_headers=http_headers,
            max_attempts=max_attempts,
            request_timeout=request_timeout,
            isolation_level=isolation_level,
            http_session=http_session,
            client_tags=client_tags,
            legacy_primitive_types=legacy_primitive_types,
            legacy_prepared_statements=legacy_prepared_statements,
            roles=roles,
            timezone=timezone,
            encoding=encoding,
        )

    @uses_trino
    @public
    @usage.method_logger
    def create_engine(
        self,
        source: str = DEFAULT_SQLALCHEMY_SOURCE,
        catalog: str = DEFAULT_CATALOG,
        schema: str = DEFAULT_SCHEMA,
        session_properties: dict | None = None,
        http_headers: dict | None = None,
        max_attempts: int = DEFAULT_MAX_ATTEMPTS,
        request_timeout: int = DEFAULT_REQUEST_TIMEOUT,
        isolation_level: IsolationLevel = AUTOCOMMIT,
        verify: bool | str = True,
        http_session: Any = None,
        client_tags: list[str] | None = None,
        legacy_primitive_types: bool = False,
        legacy_prepared_statements: bool | None = None,
        roles: dict | None = None,
        timezone: str | None = None,
        encoding: str | list[str] | None = None,
    ) -> Engine:
        """Create a SQLAlchemy engine for Trino.

        Parameters:
            source: Source identifier for Trino queries.
            catalog: Trino catalog to connect to.
            schema: Database schema within the catalog.
            session_properties: Dictionary of Trino session properties.
            http_headers: Additional HTTP headers for the connection.
            max_attempts: Maximum number of retry attempts for failed requests.
            request_timeout: Timeout in seconds for each HTTP request.
            isolation_level: Transaction isolation level.
            verify: Whether to verify SSL certificates.
                    Set verify="/path/to/cert.crt" if you want to verify the ssl cert (default: True).
            http_session: Custom HTTP session for connection pooling.
            client_tags: Tags to identify the client in Trino query logs.
            legacy_primitive_types: Whether to use legacy primitive type handling.
            legacy_prepared_statements: Whether to use legacy prepared statement handling.
            roles: Dictionary mapping catalog names to role names.
            timezone: Timezone for the session.
            encoding: Character encoding for the connection.

        Returns:
            A SQLAlchemy engine for executing queries against Trino.

        Raises:
            hopsworks_common.client.exceptions.TrinoException:
                If service discovery domain is not configured.
            hopsworks.client.exceptions.RestAPIError:
                If credentials cannot be retrieved.

        Example:
            ```python
            import hopsworks
            from sqlalchemy import text

            project = hopsworks.login()
            trino_api = project.get_trino_api()
            engine = trino_api.create_engine(catalog="iceberg", schema="my_db")
            with engine.connect() as connection:
                result = connection.execute(text("SELECT * FROM my_table"))
                for row in result:
                    print(row)
            ```
        """
        from sqlalchemy import create_engine

        host = self.get_host()
        port = self.get_port()

        user, password = self.get_basic_auth()
        basic_auth = BasicAuthentication(user, password)

        connection_url = URL(
            host=host,
            port=port,
            user=user,
            catalog=catalog,
            schema=schema,
        )
        connect_args = {
            "auth": basic_auth,
            "http_scheme": HTTPS,
            "verify": self._get_ca_chain_path(verify),
            "source": source,
            "session_properties": session_properties,
            "http_headers": http_headers,
            "client_tags": client_tags,
            "legacy_primitive_types": legacy_primitive_types,
            "legacy_prepared_statements": legacy_prepared_statements,
            "roles": roles,
            "timezone": timezone,
            "encoding": encoding,
            "max_attempts": max_attempts,
            "request_timeout": request_timeout,
            "isolation_level": isolation_level,
            "http_session": http_session,
        }
        return create_engine(connection_url, connect_args=connect_args)



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import json
from datetime import datetime, timezone
from enum import Enum
from typing import TYPE_CHECKING

import humps
from hopsworks_apigen import public
from hopsworks_common import util
from hopsworks_common.core.rest_endpoint import RestEndpointConfig
from hopsworks_common.job_schedule import JobSchedule


if TYPE_CHECKING:
    pass
else:
    from hopsworks_common.core.rest_endpoint import RestEndpointConfig


class LoadingStrategy(Enum):
    FULL_LOAD = "FULL_LOAD"
    INCREMENTAL_ID = "INCREMENTAL_ID"
    INCREMENTAL_TIMESTAMP = "INCREMENTAL_TIMESTAMP"
    INCREMENTAL_DATE = "INCREMENTAL_DATE"


class WriteMode(Enum):
    APPEND = "APPEND"
    MERGE = "MERGE"


@public("hopsworks.core.FeatureColumnMapping")
class FeatureColumnMapping:
    def __init__(self, source_column: str, feature_name: str):
        self.source_column = source_column
        self.feature_name = feature_name

    def to_dict(self):
        return {
            "sourceColumn": self.source_column,
            "featureName": self.feature_name,
        }

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    def to_json(self):
        return humps.decamelize(self.to_dict())

    @public
    @property
    def source_column(self) -> str:
        return self._source_column

    @source_column.setter
    def source_column(self, source_column: str) -> None:
        self._source_column = source_column

    @public
    @property
    def feature_name(self) -> str:
        return self._feature_name

    @feature_name.setter
    def feature_name(self, feature_name: str) -> None:
        self._feature_name = feature_name


@public("hopsworks.core.FullLoadConfig")
class FullLoadConfig:
    def __init__(
        self,
        source_cursor_field: str | None = None,
        initial_value: str | None = None,
    ):
        self._source_cursor_field = source_cursor_field
        self._initial_value = initial_value

    def to_dict(self):
        return {
            "sourceCursorField": self._source_cursor_field,
            "initialValue": self._initial_value,
        }

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(
            source_cursor_field=json_decamelized.get("source_cursor_field"),
            initial_value=json_decamelized.get("initial_value"),
        )

    @property
    def source_cursor_field(self) -> str | None:
        return self._source_cursor_field

    @source_cursor_field.setter
    def source_cursor_field(self, source_cursor_field: str | None) -> None:
        self._source_cursor_field = source_cursor_field

    @property
    def initial_value(self) -> str | None:
        return self._initial_value

    @initial_value.setter
    def initial_value(self, initial_value: str | None) -> None:
        self._initial_value = initial_value


@public("hopsworks.core.LoadingConfig")
class LoadingConfig:
    def __init__(
        self,
        loading_strategy: LoadingStrategy | str = LoadingStrategy.FULL_LOAD,
        source_cursor_field: str | None = None,
        initial_value: str | None = None,
        rest_filter_param: str | None = None,
    ):
        if isinstance(loading_strategy, LoadingStrategy):
            self._loading_strategy = loading_strategy
        elif isinstance(loading_strategy, str):
            try:
                self._loading_strategy = LoadingStrategy(loading_strategy)
            except ValueError as exc:
                valid_values = ", ".join(strategy.value for strategy in LoadingStrategy)
                raise ValueError(
                    f"Invalid loading_strategy '{loading_strategy}'. "
                    f"Valid values: {valid_values}."
                ) from exc
        else:
            raise TypeError(
                "loading_strategy must be a LoadingStrategy or str, "
                f"got {type(loading_strategy).__name__}."
            )
        self._source_cursor_field = source_cursor_field
        self._initial_value = initial_value
        self._rest_filter_param = rest_filter_param

    def to_dict(self):
        incremental_config = None
        full_load_config = None

        if self._loading_strategy in [
            LoadingStrategy.INCREMENTAL_ID,
            LoadingStrategy.INCREMENTAL_TIMESTAMP,
            LoadingStrategy.INCREMENTAL_DATE,
        ]:
            incremental_config = {"sourceCursorField": self._source_cursor_field}
            if self._rest_filter_param is not None:
                incremental_config["restFilterParam"] = self._rest_filter_param
            if self._loading_strategy in [
                LoadingStrategy.INCREMENTAL_ID,
                LoadingStrategy.INCREMENTAL_TIMESTAMP,
            ]:
                incremental_config["initialValue"] = self._initial_value
            elif self._initial_value is not None:
                if isinstance(self._initial_value, (int, float)):
                    incremental_config["initialIngestionDate"] = int(
                        self._initial_value
                    )
                else:
                    initial_value = str(self._initial_value)
                    parsed = None
                    if "T" in initial_value:
                        if initial_value.endswith("Z"):
                            parsed = datetime.fromisoformat(initial_value[:-1])
                            if parsed.tzinfo is None:
                                parsed = parsed.replace(tzinfo=timezone.utc)
                        else:
                            parsed = datetime.fromisoformat(initial_value)
                            if parsed.tzinfo is None:
                                parsed = parsed.replace(tzinfo=timezone.utc)
                    if parsed is not None:
                        incremental_config["initialIngestionDate"] = int(
                            parsed.timestamp() * 1000
                        )
                    else:
                        incremental_config["initialIngestionDate"] = (
                            util.get_timestamp_from_date_string(initial_value)
                        )
            if all(value is None for value in incremental_config.values()):
                incremental_config = None

        if self._loading_strategy == LoadingStrategy.FULL_LOAD:
            full_load_config = {
                "sourceCursorField": self._source_cursor_field,
                "initialValue": self._initial_value,
            }
            if all(value is None for value in full_load_config.values()):
                full_load_config = None

        return {
            "loadingStrategy": self._loading_strategy.value,
            "incrementalLoadingConfig": incremental_config,
            "fullLoadConfig": full_load_config,
        }

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        incremental_config = json_decamelized.get("incremental_loading_config")
        full_load_config = json_decamelized.get("full_load_config")

        source_cursor_field = None
        initial_value = None
        rest_filter_param = None
        if incremental_config:
            source_cursor_field = incremental_config.get("source_cursor_field")
            rest_filter_param = incremental_config.get("rest_filter_param")
            if "initial_value" in incremental_config:
                initial_value = incremental_config.get("initial_value")
            elif "initial_ingestion_date" in incremental_config:
                initial_value = incremental_config.get("initial_ingestion_date")
        elif full_load_config:
            source_cursor_field = full_load_config.get("source_cursor_field")
            initial_value = full_load_config.get("initial_value")

        return cls(
            loading_strategy=json_decamelized.get(
                "loading_strategy", LoadingStrategy.FULL_LOAD.value
            ),
            source_cursor_field=source_cursor_field,
            initial_value=initial_value,
            rest_filter_param=rest_filter_param,
        )

    def to_json(self):
        return humps.decamelize(self.to_dict())


@public("hopsworks.core.SinkJobConfiguration")
class SinkJobConfiguration:
    DTO_TYPE = "ingestionJobConfiguration"

    def __init__(
        self,
        name: str | None = None,
        write_mode: WriteMode | str | None = WriteMode.APPEND,
        batch_size: int | None = 100000,
        sql_source_fetch_chunk_size: int | None = 50000,
        source_read_workers: int | None = 1,
        data_processing_workers: int | None = 1,
        max_upload_batch_size_mb: int | None = 128,
        sql_table_num_partitions: int | None = 2,
        loading_config: LoadingConfig | dict | None = None,
        column_mappings: list[FeatureColumnMapping] | list[dict] | None = None,
        endpoint_config: dict | RestEndpointConfig | None = None,
        schedule_config: JobSchedule | dict | None = None,
    ):
        self._name = name
        self.write_mode = write_mode
        self._batch_size = batch_size
        self._sql_source_fetch_chunk_size = sql_source_fetch_chunk_size
        self._source_read_workers = source_read_workers
        self._data_processing_workers = data_processing_workers
        self._max_upload_batch_size_mb = max_upload_batch_size_mb
        self._sql_table_num_partitions = sql_table_num_partitions
        if isinstance(loading_config, dict):
            self._loading_config = LoadingConfig.from_response_json(loading_config)
        else:
            self._loading_config = loading_config or LoadingConfig()

        if column_mappings:
            self._column_mappings = [
                (
                    FeatureColumnMapping.from_response_json(mapping)
                    if isinstance(mapping, dict)
                    else mapping
                )
                for mapping in column_mappings
            ]
        else:
            self._column_mappings = []
        self._featuregroup_id = None
        self._featurestore_id = None
        self._storage_connector_id = None
        if isinstance(endpoint_config, dict):
            self._endpoint_config = RestEndpointConfig.from_response_json(
                endpoint_config
            )
        else:
            self._endpoint_config = endpoint_config
        self._schedule_config = (
            JobSchedule.from_response_json(schedule_config)
            if isinstance(schedule_config, dict)
            else schedule_config
        )

    def to_dict(self):
        return {
            "type": self.DTO_TYPE,
            "name": self._name,
            "writeMode": self._write_mode.value,
            "batchSize": self._batch_size,
            "sqlSourceFetchChunkSize": self._sql_source_fetch_chunk_size,
            "sourceReadWorkers": self._source_read_workers,
            "dataProcessingWorkers": self._data_processing_workers,
            "maxUploadBatchSizeMB": self._max_upload_batch_size_mb,
            "sqlTableNumPartitions": self._sql_table_num_partitions,
            "loadingConfig": (
                self._loading_config.to_dict()
                if isinstance(self._loading_config, LoadingConfig)
                else self._loading_config
            ),
            "columnMappings": [
                self._normalize_column_mapping(mapping)
                for mapping in self._column_mappings
            ],
            "featuregroupId": self._featuregroup_id,
            "featurestoreId": self._featurestore_id,
            "storageConnectorId": self._storage_connector_id,
            "endpointConfig": (
                self._endpoint_config.to_dict()
                if hasattr(self._endpoint_config, "to_dict")
                else self._endpoint_config
            ),
            "jobSchedule": (
                self._schedule_config.to_dict()
                if isinstance(self._schedule_config, JobSchedule)
                else self._schedule_config
            ),
        }

    @staticmethod
    def _normalize_column_mapping(mapping):
        if isinstance(mapping, FeatureColumnMapping):
            return mapping.to_dict()
        if isinstance(mapping, dict):
            if "sourceColumn" in mapping and "featureName" in mapping:
                return mapping
            if "source_column" in mapping and "feature_name" in mapping:
                return {
                    "sourceColumn": mapping["source_column"],
                    "featureName": mapping["feature_name"],
                }
        source_column = getattr(mapping, "source_column", None)
        feature_name = getattr(mapping, "feature_name", None)
        if source_column is not None and feature_name is not None:
            return {"sourceColumn": source_column, "featureName": feature_name}
        return mapping

    def json(self):
        return json.dumps(self.to_dict())

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        loading_config = LoadingConfig.from_response_json(
            json_decamelized.get("loading_config", {})
        )
        column_mappings = [
            FeatureColumnMapping.from_response_json(mapping)
            for mapping in json_decamelized.get("column_mappings", [])
        ]
        job_schedule = json_decamelized.get("job_schedule", None)
        endpoint_config = json_dict.get("endpointConfig", None)
        return SinkJobConfiguration(
            write_mode=json_decamelized.get("write_mode", WriteMode.APPEND.value),
            batch_size=json_decamelized.get("batch_size", 100000),
            sql_source_fetch_chunk_size=json_decamelized.get(
                "sql_source_fetch_chunk_size", 50000
            ),
            source_read_workers=json_decamelized.get("source_read_workers", 1),
            data_processing_workers=json_decamelized.get("data_processing_workers", 1),
            max_upload_batch_size_mb=json_decamelized.get(
                "max_upload_batch_size_mb", 128
            ),
            sql_table_num_partitions=json_decamelized.get(
                "sql_table_num_partitions", 2
            ),
            name=json_decamelized.get("name", None),
            loading_config=loading_config,
            column_mappings=column_mappings,
            endpoint_config=(
                RestEndpointConfig.from_response_json(endpoint_config)
                if endpoint_config
                else None
            ),
            schedule_config=(
                JobSchedule.from_response_json(job_schedule) if job_schedule else None
            ),
        )

    def set_extra_params(self, **kwargs) -> None:
        self._featuregroup_id = kwargs.get("featuregroup_id")
        self._featurestore_id = kwargs.get("featurestore_id")
        self._storage_connector_id = kwargs.get("storage_connector_id")
        endpoint_config = kwargs.get("endpoint_config")
        if isinstance(endpoint_config, dict):
            self._endpoint_config = RestEndpointConfig.from_response_json(
                endpoint_config
            )
        else:
            self._endpoint_config = endpoint_config
        self._name = kwargs.get("name", self._name)

    @public
    @property
    def batch_size(self) -> int | None:
        return self._batch_size

    @batch_size.setter
    def batch_size(self, batch_size: int | None) -> None:
        self._batch_size = batch_size

    @public
    @property
    def sql_source_fetch_chunk_size(self) -> int | None:
        return self._sql_source_fetch_chunk_size

    @sql_source_fetch_chunk_size.setter
    def sql_source_fetch_chunk_size(
        self, sql_source_fetch_chunk_size: int | None
    ) -> None:
        self._sql_source_fetch_chunk_size = sql_source_fetch_chunk_size

    @property
    def source_read_workers(self) -> int | None:
        return self._source_read_workers

    @source_read_workers.setter
    def source_read_workers(self, source_read_workers: int | None) -> None:
        self._source_read_workers = source_read_workers

    @property
    def data_processing_workers(self) -> int | None:
        return self._data_processing_workers

    @data_processing_workers.setter
    def data_processing_workers(self, data_processing_workers: int | None) -> None:
        self._data_processing_workers = data_processing_workers

    @property
    def max_upload_batch_size_mb(self) -> int | None:
        return self._max_upload_batch_size_mb

    @max_upload_batch_size_mb.setter
    def max_upload_batch_size_mb(self, max_upload_batch_size_mb: int | None) -> None:
        self._max_upload_batch_size_mb = max_upload_batch_size_mb

    @property
    def sql_table_num_partitions(self) -> int | None:
        return self._sql_table_num_partitions

    @sql_table_num_partitions.setter
    def sql_table_num_partitions(self, sql_table_num_partitions: int | None) -> None:
        self._sql_table_num_partitions = sql_table_num_partitions

    @property
    def loading_config(self) -> LoadingConfig | dict | None:
        return self._loading_config

    @loading_config.setter
    def loading_config(self, loading_config: LoadingConfig | dict | None) -> None:
        self._loading_config = loading_config

    @public
    @property
    def column_mappings(self) -> list[FeatureColumnMapping] | list[dict] | None:
        return self._column_mappings

    @column_mappings.setter
    def column_mappings(
        self, column_mappings: list[FeatureColumnMapping] | list[dict] | None
    ) -> None:
        self._column_mappings = column_mappings

    @public
    @property
    def name(self) -> str | None:
        return self._name

    @name.setter
    def name(self, name: str | None) -> None:
        self._name = name

    @public
    @property
    def write_mode(self) -> WriteMode:
        return self._write_mode

    @write_mode.setter
    def write_mode(self, write_mode: WriteMode | str | None) -> None:
        if write_mode is None:
            self._write_mode = WriteMode.APPEND
        elif isinstance(write_mode, WriteMode):
            self._write_mode = write_mode
        elif isinstance(write_mode, str):
            try:
                self._write_mode = WriteMode(write_mode.upper())
            except ValueError as exc:
                valid_values = ", ".join(mode.value for mode in WriteMode)
                raise ValueError(
                    f"Invalid write_mode '{write_mode}'. Valid values: {valid_values}."
                ) from exc
        else:
            raise TypeError(
                "write_mode must be a WriteMode or str, "
                f"got {type(write_mode).__name__}."
            )

    @property
    def schedule_config(self) -> JobSchedule | None:
        return self._schedule_config

    @schedule_config.setter
    def schedule_config(self, schedule_config: JobSchedule | dict | None) -> None:
        self._schedule_config = (
            JobSchedule.from_response_json(schedule_config)
            if isinstance(schedule_config, dict)
            else schedule_config
        )



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from enum import Enum
from typing import Any

import humps
from hopsworks_apigen import public
from hopsworks_common import util


class PaginationType(str, Enum):
    JSON_LINK = "json_link"
    HEADER_LINK = "header_link"
    HEADER_CURSOR = "header_cursor"
    OFFSET = "offset"
    PAGE_NUMBER = "page_number"
    CURSOR = "cursor"
    SINGLE_PAGE = "single_page"
    AUTO = "auto"


@public("hopsworks.core.rest_endpoint.QueryParams")
class QueryParams:
    """Represents an arbitrary name/value pair used by REST endpoint configurations."""

    def __init__(self, name: str, value: Any) -> None:
        self._name = name
        self._value = value

    @classmethod
    def from_response_json(cls, json_dict: dict[str, Any]) -> QueryParams | None:
        if json_dict is None:
            return None
        json_decamelized: dict[str, Any] = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    def to_dict(self) -> dict[str, Any]:
        return {
            "name": self._name,
            "value": self._value,
        }

    def json(self) -> str:
        return json.dumps(humps.decamelize(self.to_dict()), cls=util.Encoder)

    def __repr__(self) -> str:
        return f"QueryParams({self._name!r}, {self._value!r})"

    @public
    @property
    def name(self) -> str:
        return self._name

    @property
    def key(self) -> str:
        return self._name

    @public
    @property
    def value(self) -> Any:
        return self._value


@public("hopsworks.core.rest_endpoint.PaginationConfig")
class PaginationConfig:
    """Base pagination configuration that specific strategies inherit from."""

    pagination_type: PaginationType | None = None

    def __init__(self, **extra_params: Any) -> None:
        self._extra_params = {
            key: value
            for key, value in (extra_params or {}).items()
            if value is not None
        }

    @classmethod
    def from_type(
        cls,
        pagination_type: str | PaginationType | None,
        **kwargs,
    ) -> PaginationConfig | None:
        if pagination_type is None:
            return None

        if isinstance(pagination_type, PaginationType):
            pagination_enum = pagination_type
        elif isinstance(pagination_type, str):
            pagination_enum = PaginationType(pagination_type.lower())
        else:
            raise TypeError(
                "pagination_type must be a PaginationType or str, "
                f"got {type(pagination_type).__name__}."
            )
        pagination_class = _PAGINATION_CLASS_MAP.get(pagination_enum)
        if pagination_class is None:
            raise ValueError(
                f"Unsupported pagination type '{pagination_enum.value}'. "
                f"Supported types: {[ptype.value for ptype in _PAGINATION_CLASS_MAP]}"
            )
        return pagination_class(**kwargs)

    @classmethod
    def from_response_json(cls, json_dict: dict[str, Any]) -> PaginationConfig | None:
        if json_dict is None:
            return None

        json_decamelized: dict[str, Any] = humps.decamelize(json_dict)
        pagination_type = json_decamelized.pop("type", None)
        if pagination_type is None:
            raise ValueError("Pagination configuration must include the 'type' field.")
        return cls.from_type(pagination_type, **json_decamelized)

    def to_dict(self) -> dict[str, Any]:
        payload = self._merge_payload(self._payload())
        payload_camelized = humps.camelize(payload) if payload else {}

        if self.pagination_type is not None:
            payload_camelized["type"] = self.pagination_type.value.upper()
        return payload_camelized

    def json(self) -> str:
        return json.dumps(humps.decamelize(self.to_dict()), cls=util.Encoder)

    def __repr__(self) -> str:
        payload = self._merge_payload(self._payload())
        pagination_type = self.pagination_type.value if self.pagination_type else None
        return (
            f"{self.__class__.__name__}(type={pagination_type!r}, config={payload!r})"
        )

    def _payload(self) -> dict[str, Any]:
        """Return pagination-specific options. Subclasses should override."""
        return {}

    def _merge_payload(self, payload: dict[str, Any]) -> dict[str, Any]:
        merged = {key: value for key, value in payload.items() if value is not None}
        for key, value in self._extra_params.items():
            merged.setdefault(key, value)
        return merged


@public("hopsworks.core.rest_endpoint.JsonLinkPaginationConfig")
class JsonLinkPaginationConfig(PaginationConfig):
    pagination_type = PaginationType.JSON_LINK

    def __init__(self, next_url_path: str, **kwargs) -> None:
        if not next_url_path:
            raise ValueError("JsonLinkPaginationConfig requires 'next_url_path'.")
        super().__init__(**kwargs)
        self._next_url_path = next_url_path

    def _payload(self) -> dict[str, Any]:
        return {
            "next_url_path": self._next_url_path,
        }


@public("hopsworks.core.rest_endpoint.HeaderLinkPaginationConfig")
class HeaderLinkPaginationConfig(PaginationConfig):
    pagination_type = PaginationType.HEADER_LINK

    def __init__(self, links_next_key: str = "next", **kwargs) -> None:
        super().__init__(**kwargs)
        self._links_next_key = links_next_key or "next"

    def _payload(self) -> dict[str, Any]:
        return {
            "links_next_key": self._links_next_key,
        }


@public("hopsworks.core.rest_endpoint.HeaderCursorPaginationConfig")
class HeaderCursorPaginationConfig(PaginationConfig):
    pagination_type = PaginationType.HEADER_CURSOR

    def __init__(
        self,
        cursor_key: str = "next",
        cursor_param: str = "cursor",
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self._cursor_key = cursor_key or "next"
        self._cursor_param = cursor_param or "cursor"

    def _payload(self) -> dict[str, Any]:
        return {
            "cursor_key": self._cursor_key,
            "cursor_param": self._cursor_param,
        }


@public("hopsworks.core.rest_endpoint.OffsetPaginationConfig")
class OffsetPaginationConfig(PaginationConfig):
    pagination_type = PaginationType.OFFSET

    def __init__(
        self,
        limit: int | None = None,
        offset: int = 0,
        offset_param: str = "offset",
        offset_body_path: str | None = None,
        limit_param: str = "limit",
        limit_body_path: str | None = None,
        total_path: str | None = None,
        maximum_offset: int | None = None,
        stop_after_empty_page: bool = True,
        has_more_path: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self._limit = limit
        self._offset = offset
        self._offset_param = offset_param or "offset"
        self._offset_body_path = offset_body_path
        self._limit_param = limit_param or "limit"
        self._limit_body_path = limit_body_path
        self._total_path = total_path
        self._maximum_offset = maximum_offset
        self._stop_after_empty_page = stop_after_empty_page
        self._has_more_path = has_more_path

    def _payload(self) -> dict[str, Any]:
        return {
            "limit": self._limit,
            "offset": self._offset,
            "offset_param": self._offset_param,
            "offset_body_path": self._offset_body_path,
            "limit_param": self._limit_param,
            "limit_body_path": self._limit_body_path,
            "total_path": self._total_path,
            "maximum_offset": self._maximum_offset,
            "stop_after_empty_page": self._stop_after_empty_page,
            "has_more_path": self._has_more_path,
        }


@public("hopsworks.core.rest_endpoint.PageNumberPaginationConfig")
class PageNumberPaginationConfig(PaginationConfig):
    pagination_type = PaginationType.PAGE_NUMBER

    def __init__(
        self,
        base_page: int = 0,
        page_param: str = "page",
        total_path: str | None = None,
        maximum_page: int | None = None,
        stop_after_empty_page: bool = True,
        has_more_path: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self._base_page = base_page
        self._page_param = page_param or "page"
        self._total_path = total_path
        self._maximum_page = maximum_page
        self._stop_after_empty_page = stop_after_empty_page
        self._has_more_path = has_more_path

    def _payload(self) -> dict[str, Any]:
        return {
            "base_page": self._base_page,
            "page_param": self._page_param,
            "total_path": self._total_path,
            "maximum_page": self._maximum_page,
            "stop_after_empty_page": self._stop_after_empty_page,
            "has_more_path": self._has_more_path,
        }


@public("hopsworks.core.rest_endpoint.CursorPaginationConfig")
class CursorPaginationConfig(PaginationConfig):
    pagination_type = PaginationType.CURSOR

    def __init__(
        self,
        cursor_path: str = "cursors.next",
        cursor_param: str | None = None,
        cursor_body_path: str | None = None,
        **kwargs,
    ) -> None:
        if cursor_param and cursor_body_path:
            raise ValueError(
                "CursorPaginationConfig cannot set both 'cursor_param' and 'cursor_body_path'."
            )
        super().__init__(**kwargs)
        self._cursor_path = cursor_path or "cursors.next"
        if cursor_param is None and cursor_body_path is None:
            cursor_param = "cursor"
        self._cursor_param = cursor_param
        self._cursor_body_path = cursor_body_path

    def _payload(self) -> dict[str, Any]:
        return {
            "cursor_path": self._cursor_path,
            "cursor_param": self._cursor_param,
            "cursor_body_path": self._cursor_body_path,
        }


@public("hopsworks.core.rest_endpoint.SinglePagePaginationConfig")
class SinglePagePaginationConfig(PaginationConfig):
    pagination_type = PaginationType.SINGLE_PAGE

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)


@public("hopsworks.core.rest_endpoint.AutoPaginationConfig")
class AutoPaginationConfig(PaginationConfig):
    pagination_type = PaginationType.AUTO

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)


_PAGINATION_CLASS_MAP: dict[PaginationType, type[PaginationConfig]] = {
    PaginationType.JSON_LINK: JsonLinkPaginationConfig,
    PaginationType.HEADER_LINK: HeaderLinkPaginationConfig,
    PaginationType.HEADER_CURSOR: HeaderCursorPaginationConfig,
    PaginationType.OFFSET: OffsetPaginationConfig,
    PaginationType.PAGE_NUMBER: PageNumberPaginationConfig,
    PaginationType.CURSOR: CursorPaginationConfig,
    PaginationType.SINGLE_PAGE: SinglePagePaginationConfig,
    PaginationType.AUTO: AutoPaginationConfig,
}


@public("hopsworks.core.rest_endpoint.RestEndpointConfig")
class RestEndpointConfig:
    """Configuration describing how a REST resource should be accessed."""

    def __init__(
        self,
        relative_url: str | None = None,
        query_params: list[QueryParams] | dict[str, Any] | None = None,
        pagination_config: PaginationConfig | dict[str, Any] | None = None,
        **kwargs,
    ) -> None:
        self._relative_url = relative_url
        self._query_params = self._init_query_params(query_params)
        self._pagination_config = self._init_pagination_config(pagination_config)

    @classmethod
    def from_response_json(cls, json_dict: dict[str, Any]) -> RestEndpointConfig | None:
        if json_dict is None:
            return None

        json_decamelized: dict[str, Any] = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    def to_dict(self) -> dict[str, Any]:
        payload: dict[str, Any] = {}
        if self._relative_url is not None:
            payload["relativeUrl"] = self._relative_url
        if self._query_params:
            payload["queryParams"] = [param.to_dict() for param in self._query_params]
        if self._pagination_config is not None:
            payload["paginationConfig"] = self._pagination_config.to_dict()
        return payload

    def json(self) -> str:
        return json.dumps(humps.decamelize(self.to_dict()), cls=util.Encoder)

    def __repr__(self) -> str:
        return (
            "RestEndpointConfig("
            f"relative_url={self._relative_url!r}, "
            f"query_params={self._query_params!r}, "
            f"pagination_config={self._pagination_config!r}"
            ")"
        )

    def _init_query_params(
        self, query_params: list[QueryParams] | dict[str, Any] | None
    ) -> list[QueryParams]:
        if query_params is None:
            return []

        if isinstance(query_params, dict):
            return [
                QueryParams(name=key, value=value)
                for key, value in query_params.items()
            ]

        if isinstance(query_params, list):
            normalized: list[QueryParams] = []
            for param in query_params:
                if isinstance(param, QueryParams):
                    normalized.append(param)
                elif isinstance(param, dict):
                    normalized.append(QueryParams(**param))
                else:
                    raise TypeError(
                        "Query parameters must be provided as QueryParams instances or dictionaries."
                    )
            return normalized

        raise TypeError(
            "query_params must be a list of QueryParams, a dictionary, or None."
        )

    def _init_pagination_config(
        self, pagination_config: PaginationConfig | dict[str, Any] | None
    ) -> PaginationConfig | None:
        if pagination_config is None:
            return None
        if isinstance(pagination_config, PaginationConfig):
            return pagination_config
        if isinstance(pagination_config, str):
            return PaginationConfig.from_type(pagination_config)
        if isinstance(pagination_config, dict):
            return PaginationConfig.from_response_json(pagination_config)
        raise TypeError(
            "pagination_config must be a PaginationConfig or a dictionary representation."
        )

    @public
    @property
    def relative_url(self) -> str | None:
        return self._relative_url

    @public
    @property
    def query_params(self) -> list[QueryParams]:
        return self._query_params

    @public
    @property
    def pagination_config(self) -> PaginationConfig | None:
        return self._pagination_config



# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

# This implementation has been borrowed from kserve/kserve repository
# https://github.com/kserve/kserve/blob/release-0.11/python/kserve/kserve/exceptions.py
from __future__ import annotations

import six
from hopsworks_apigen import also_available_as


@also_available_as("hsml.client.istio.grpc.exceptions.OpenApiException")
class OpenApiException(Exception):
    """The base exception class for all OpenAPIExceptions."""


@also_available_as("hsml.client.istio.grpc.exceptions.ApiTypeError")
class ApiTypeError(OpenApiException, TypeError):
    def __init__(
        self,
        msg: str,
        path_to_item: list | None = None,
        valid_classes: tuple | None = None,
        key_type: bool | None = None,
    ) -> None:
        """Raises an exception for TypeErrors.

        Parameters:
            msg: the exception message
            path_to_item: a list of keys an indices to get to the current_item; None if unset
            valid_classes: the primitive classes that current item should be an instance of; None if unset
            key_type: False if our value is a value in a dict; True if it is a key in a dict; False if our item is an item in a list; None if unset
        """
        self.path_to_item = path_to_item
        self.valid_classes = valid_classes
        self.key_type = key_type
        full_msg = msg
        if path_to_item:
            full_msg = f"{msg} at {render_path(path_to_item)}"
        super().__init__(full_msg)


@also_available_as("hsml.client.istio.grpc.exceptions.ApiValueError")
class ApiValueError(OpenApiException, ValueError):
    def __init__(self, msg: str, path_to_item: list | None = None):
        """An error related to API value.

        Parameters:
            msg: the exception message.
            path_to_item: the path to the exception in the received_data dict.
        """
        self.path_to_item = path_to_item
        full_msg = msg
        if path_to_item:
            full_msg = f"{msg} at {render_path(path_to_item)}"
        super().__init__(full_msg)


@also_available_as("hsml.client.istio.grpc.exceptions.ApiKeyError")
class ApiKeyError(OpenApiException, KeyError):
    def __init__(self, msg: str, path_to_item: list | None = None):
        """An error related to the API key.

        Parameters:
            msg: the exception message.
            path_to_item:
                the path to the exception in the received_data dict
        """
        self.path_to_item = path_to_item
        full_msg = msg
        if path_to_item:
            full_msg = f"{msg} at {render_path(path_to_item)}"
        super().__init__(full_msg)


@also_available_as("hsml.client.istio.grpc.exceptions.ApiException")
class ApiException(OpenApiException):
    def __init__(self, status=None, reason=None, http_resp=None):
        if http_resp:
            self.status = http_resp.status
            self.reason = http_resp.reason
            self.body = http_resp.data
            self.headers = http_resp.getheaders()
        else:
            self.status = status
            self.reason = reason
            self.body = None
            self.headers = None

    def __str__(self):
        """Custom error messages for exception."""
        error_message = f"({self.status})\nReason: {self.reason}\n"
        if self.headers:
            error_message += f"HTTP response headers: {self.headers}\n"

        if self.body:
            error_message += f"HTTP response body: {self.body}\n"

        return error_message


@also_available_as("hsml.client.istio.grpc.exceptions.render_path")
def render_path(path_to_item: list) -> str:
    """Returns a string representation of a path.

    Parameters:
        path_to_item: A list of path parts.

    Returns:
        A string representation of the path.
    """
    result = ""
    for pth in path_to_item:
        if isinstance(pth, six.integer_types):
            result += f"[{pth}]"
        else:
            result += f"['{pth}']"
    return result



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json

from hopsworks_apigen import also_available_as
from hopsworks_common import client, library


@also_available_as("hopsworks.core.library_api.LibraryApi")
class LibraryApi:
    def _install(
        self, library_name: str, name: str, library_spec: dict
    ) -> library.Library:
        """Install a library in the environment.

        Parameters:
            library_name: Name of the library.
            name: Name of the environment.
            library_spec: Installation payload.

        Returns:
            The library object.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        _client = client.get_instance()

        path_params = [
            "project",
            _client._project_id,
            "python",
            "environments",
            name,
            "libraries",
            library_name,
        ]

        headers = {"content-type": "application/json"}
        return library.Library.from_response_json(
            _client._send_request(
                "POST", path_params, headers=headers, data=json.dumps(library_spec)
            ),
            environment=self,
        )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Literal

import humps
from hopsworks_apigen import public
from hopsworks_common import alert, client, util
from hopsworks_common.core import (
    alerts_api,
    dataset_api,
    environment_api,
    flink_cluster_api,
    git_api,
    job_api,
    kafka_api,
    opensearch_api,
    search_api,
    trino_api,
)


if TYPE_CHECKING:
    from hsfs.feature_store import FeatureStore
    from hsml.model_registry import ModelRegistry
    from hsml.model_serving import ModelServing


@public("hopsworks.project.Project")
class Project:
    """Class representing a Hopsworks project.

    Use [`hopsworks.login`][hopsworks.login] to get the current project after logging in.

    Use [`hopsworks.create_project`][hopsworks.create_project] to create a new project and get the project object.
    """

    def __init__(
        self,
        archived=None,
        created=None,
        description=None,
        docker_image=None,
        hops_examples=None,
        inodeid=None,
        is_old_docker_image=None,
        is_preinstalled_docker_image=None,
        owner=None,
        project_id=None,
        project_name=None,
        project_team=None,
        quotas=None,
        retention_period=None,
        services=None,
        datasets=None,
        creation_status=None,
        project_namespace=None,
        **kwargs,
    ):
        self._id = project_id
        self._name = project_name
        self._owner = owner
        self._description = description
        self._created = created

        self._opensearch_api = opensearch_api.OpenSearchApi()
        self._kafka_api = kafka_api.KafkaApi()
        self._job_api = job_api.JobApi()
        self._jobs_api = self._job_api  # deprecated
        self._flink_cluster_api = flink_cluster_api.FlinkClusterApi()
        self._git_api = git_api.GitApi()
        self._dataset_api = dataset_api.DatasetApi()
        self._environment_api = environment_api.EnvironmentApi()
        self._alerts_api = alerts_api.AlertsApi()
        self._search_api = search_api.SearchApi()
        self._project_namespace = project_namespace
        self._trino_api = None

    @classmethod
    def from_response_json(cls, json_dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        return None

    @public
    @property
    def id(self):
        """Id of the project."""
        return self._id

    @public
    @property
    def name(self):
        """Name of the project."""
        return self._name

    @public
    @property
    def owner(self):
        """Owner of the project."""
        return self._owner

    @public
    @property
    def description(self):
        """Description of the project."""
        return self._description

    @public
    @property
    def created(self):
        """Timestamp when the project was created."""
        return self._created

    @public
    @property
    def project_namespace(self):
        """Kubernetes namespace used by project."""
        return self._project_namespace

    @public
    def get_feature_store(self, name: str | None = None) -> FeatureStore:
        """Connect to Project's Feature Store.

        Defaulting to the project name of default feature store. To get a
        shared feature store, the project name of the feature store is required.

        Example: Example for getting the Feature Store API of a project
            ```python
            import hopsworks

            project = hopsworks.login()

            fs = project.get_feature_store()
            ```

        Parameters:
            name: Project name of the feature store.

        Returns:
            The Feature Store API.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return client.get_connection().get_feature_store(name)

    @public
    def get_model_registry(self) -> ModelRegistry:
        """Connect to Project's Model Registry API.

        Example: Example for getting the Model Registry API of a project
            ```python
            import hopsworks

            project = hopsworks.login()

            mr = project.get_model_registry()
            ```

        Returns:
            The Model Registry API.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return client.get_connection().get_model_registry()

    @public
    def get_model_serving(self) -> ModelServing:
        """Connect to Project's Model Serving API.

        Example: Example for getting the Model Serving API of a project
            ```python
            import hopsworks

            project = hopsworks.login()

            ms = project.get_model_serving()
            ```

        Returns:
            The Model Serving API.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return client.get_connection().get_model_serving()

    @public
    def get_kafka_api(self) -> kafka_api.KafkaApi:
        """Get the kafka api for the project.

        Returns:
            The Kafka Api handle.
        """
        _client = client.get_instance()
        if _client._is_external():
            _client.download_certs()
        return self._kafka_api

    @public
    def get_opensearch_api(self) -> opensearch_api.OpenSearchApi:
        """Get the opensearch api for the project.

        Returns:
            The OpenSearch Api handle.
        """
        _client = client.get_instance()
        if _client._is_external():
            _client.download_certs()
        return self._opensearch_api

    @public
    def get_job_api(self) -> job_api.JobApi:
        """Get the job API for the project.

        Returns:
            The Job Api handle.
        """
        return self._job_api

    def get_jobs_api(self):
        """**Deprecated**, use get_job_api instead. Excluded from docs to prevent API breakage."""
        return self.get_job_api()

    @public
    def get_flink_cluster_api(self) -> flink_cluster_api.FlinkClusterApi:
        """Get the flink cluster API for the project.

        Returns:
            The Flink Cluster Api handle.
        """
        return self._flink_cluster_api

    @public
    def get_git_api(self) -> git_api.GitApi:
        """Get the git repository api for the project.

        Returns:
            The Git Api handle.
        """
        return self._git_api

    @public
    def get_dataset_api(self) -> dataset_api.DatasetApi:
        """Get the dataset api for the project.

        Returns:
            The Datasets Api handle.
        """
        return self._dataset_api

    @public
    def get_environment_api(self) -> environment_api.EnvironmentApi:
        """Get the Python environment API for the project.

        Returns:
            The Python Environment Api handle.
        """
        return self._environment_api

    @public
    def get_alerts_api(self) -> alerts_api.AlertsApi:
        """Get the alerts api for the project.

        Returns:
            The Alerts Api handle.
        """
        return self._alerts_api

    @public
    def get_search_api(self) -> search_api.SearchApi:
        """Get the search api for the project.

        Returns:
            The Search Api handle.
        """
        return self._search_api

    @public
    def get_trino_api(self) -> trino_api.TrinoApi:
        """Get the Trino API for the project.

        Returns:
            The Trino API handle.
        """
        if self._trino_api is None:
            self._trino_api = trino_api.TrinoApi(project=self)
        return self._trino_api

    @public
    def get_alerts(self) -> list[alert.ProjectAlert]:
        """Get all alerts for the project.

        Returns:
            List of `ProjectAlert` objects.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return self._alerts_api.get_alerts()

    @public
    def get_alert(self, alert_id: int) -> alert.ProjectAlert | None:
        """Get an alert for the project by ID.

        Parameters:
            alert_id: The ID of the alert.

        Returns:
            The ProjectAlert object.
        """
        return self._alerts_api.get_alert(alert_id)

    @public
    def create_job_alert(
        self,
        receiver: str,
        status: Literal["job_finished", "job_failed", "job_killed", "job_long_running"],
        severity: Literal["critical", "warning", "info"],
    ) -> alert.ProjectAlert:
        """Create an alert for jobs in this project.

        Example: Example for creating a job alert
            ```python
            import hopsworks
            project = hopsworks.login()
            project.create_job_alert("my_receiver", "long_running", "info")
            ```

        Parameters:
            receiver: The receiver of the alert.
            status: The status of the alert.
            severity: The severity of the alert.

        Returns:
            The created `ProjectAlert` object.

        Raises:
            ValueError: If `status` or `severity` is invalid, also if `receiver` is `None`.
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return self._alerts_api.create_project_alert(receiver, status, severity, "Jobs")

    @public
    def create_featurestore_alert(
        self,
        receiver: str,
        status: Literal[
            "feature_validation_success",
            "feature_validation_warning",
            "feature_validation_failure",
            "feature_monitor_shift_undetected",
            "feature_monitor_shift_detected",
        ],
        severity: Literal["critical", "warning", "info"],
    ) -> alert.ProjectAlert:
        """Create an alert for feature validation and monitoring in this project.

        Example: Example for creating a featurestore alert
            ```python
            import hopsworks
            project = hopsworks.login()
            project.create_featurestore_alert("my_receiver", "feature_validation_success", "info")
            ```

        Parameters:
            receiver: The receiver of the alert.
            status: The status of the alert.
            severity: The severity of the alert.

        Returns:
            The created `ProjectAlert` object.

        Raises:
            ValueError: If `status` or `severity` is invalid, also if `receiver` is `None`.
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request.
        """
        return self._alerts_api.create_project_alert(
            receiver, status, severity, "Featurestore"
        )

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        if self._description is not None:
            return f"Project({self._name!r}, {self._owner!r}, {self._description!r})"
        return f"Project({self._name!r}, {self._owner!r})"

    @public
    def get_url(self):
        """Get url to the project in Hopsworks."""
        path = "/p/" + str(self.id)
        return util.get_hostname_replaced_url(path)



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json

import humps
from hopsworks_apigen import public
from hopsworks_common import usage, util
from hopsworks_common.core import git_remote_api


@public("hopsworks.git_remote.GitRemote")
class GitRemote:
    def __init__(
        self,
        remote_name=None,
        remote_url=None,
        url=None,
        name=None,
        type=None,
        href=None,
        expand=None,
        items=None,
        count=None,
        **kwargs,
    ):
        self._name = remote_name
        self._url = remote_url

        self._git_remote_api = git_remote_api.GitRemoteApi()

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        if "count" in json_decamelized:
            if json_decamelized["count"] == 0:
                return []
            return [cls(**remote) for remote in json_decamelized["items"]]
        return cls(**json_decamelized)

    @public
    @property
    def name(self):
        """Name of the remote."""
        return self._name

    @public
    @property
    def url(self):
        """Url of the remote."""
        return self._url

    @public
    @usage.method_logger
    def delete(self):
        """Remove the git remote from the repo.

        Raises:
            hopsworks.client.exceptions.RestAPIError: If the backend encounters an error when handling the request
        """
        self._git_remote_api._delete(self._repo_id, self.name)

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"GitRemote({self._name!r}, {self._url!r})"



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import asyncio
import base64
import itertools
import json
import logging
import threading
import traceback
import uuid
from datetime import date, datetime, timezone
from typing import TYPE_CHECKING, Any, Callable

from hopsworks_apigen import public
from hsfs.core.feature_logging_client import (
    get_instance as get_feature_logging_client,
)
from hsfs.core.feature_logging_client import (
    init_client,
)
from hsfs.core.kafka_engine import encode_row, get_writer_function
from hsfs.feature_logger import FeatureLogger


if TYPE_CHECKING:
    from hsfs.feature_view import FeatureView


_logger = logging.getLogger(__name__)


class EventEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bytes):
            # Convert bytes to a base64-encoded string
            return base64.b64encode(obj).decode("utf-8")
        if isinstance(obj, datetime):
            obj = obj.replace(tzinfo=timezone.utc)
            return obj.isoformat()
        if isinstance(obj, date):
            # Convert to days since Unix epoch
            epoch = datetime.date(1970, 1, 1)
            return (obj - epoch).days
        return None


class AsyncWorkerThread(threading.Thread):
    """Thread class to run an asyncio event loop in a separate thread. The event loop is used to run async workers that processes logs."""

    def __init__(
        self, group=None, target=None, name=None, args=..., kwargs=None, *, daemon=None
    ):
        super().__init__(group, target, name, args, kwargs, daemon=daemon)
        self._event_loop = asyncio.new_event_loop()
        self._workers = []  # List to keep track of worker coroutines
        self._tasks_queue = asyncio.Queue()  # Queue to manage tasks

        self._stop_event = (
            threading.Event()
        )  # Stop event to stop input of new tasks after a close has been called.

    def submit_task(self, task: tuple[dict, dict]):
        """Function to submit a task to the queue from a different thread so that it can be processed by workers.

        Parameters:
            task: Tuple that contains untransformed and transformed features to be logged.
        """
        if self._stop_event.is_set():
            _logger.error("Cannot submit task. Workers are stopped.")
        else:
            asyncio.run_coroutine_threadsafe(
                self._tasks_queue.put(task), self._event_loop
            )

    def _initialize_workers(self, num_workers: int, worker_function: Callable):
        """Function to initialize workers as tasks in the event loop.

        Parameters:
            num_worker: Number of workers to be initialized.
            worker_function: Function to be run by the workers.
        """
        for _ in range(num_workers):
            worker = self._event_loop.create_task(self._worker(worker_function))
            self._workers.append(worker)

    def run(self):
        """Thread functions that runs the event loop in the thread and close the event loop with feature logging after the loop is stopped."""
        asyncio.set_event_loop(self._event_loop)

        # Run the event loop
        self._event_loop.run_forever()

        # Closing the feature logging client inside the thread.
        self._event_loop.run_until_complete(get_feature_logging_client().close())

        # Close the event loop
        self._event_loop.close()

    async def _worker(self, worker_function: Callable):
        """Function to run the worker function in the event loop, until a None has been submitted to the queue.

        Parameters:
            worker_function: Function to be run by the workers.
        """
        while True:
            task = await self._tasks_queue.get()
            if task is None:
                # Poison pill means shutdown
                self._tasks_queue.task_done()
                break
            await worker_function(task)
            self._tasks_queue.task_done()

    def close(self):
        """Function to stop any more tasks from being submitted and start the graceful stop of the thread."""
        # Stop any more tasks from being submitted using the stop event.
        self._stop_event.set()

        # Stop the event loop
        asyncio.run_coroutine_threadsafe(self.finalize_event_loop(), self._event_loop)

    async def finalize_event_loop(self):
        """Function that gracefully stops the event loop by stopping the workers and waiting for all tasks to be processed."""
        # Stop workers
        for _ in range(len(self._workers)):
            await self._tasks_queue.put(None)  # Poison pill to stop workers

        # Wait until all tasks in the queue are processed
        await self._tasks_queue.join()

        # Wait until all tasks in the queue are processed
        await asyncio.gather(*self._workers)

        # Stop the event loop
        self._event_loop.stop()


@public
class AsyncFeatureLogger(FeatureLogger):
    def __init__(
        self,
        project_id,
        source,
        namespace,
        deployment_name,
        max_concurrent_tasks=5,
        feature_logger_config: dict[str, Any] | None = None,
    ):
        self._max_concurrent_tasks = max_concurrent_tasks
        self._feature_view: FeatureView = None
        self._project_id = project_id
        self._source = source
        self._namespace = namespace
        self._deployment_name = deployment_name
        self._workers = []  # List to keep track of worker coroutines

        # Initialize workers in another so that we don't cause any issues with the event loop's running in the main thread.
        self._async_worker_thread = AsyncWorkerThread()

        self._feature_logger_config = feature_logger_config
        if self._feature_logger_config is None:
            self._feature_logger_config = {}
        # set pool size equals to the number of concurrent tasks
        if "pool_size" not in self._feature_logger_config:
            self._feature_logger_config["pool_size"] = max_concurrent_tasks
        self._feature_encoders = {}

    @public
    def log(
        self,
        untransformed_features: list[dict] = None,
        transformed_features: list[dict] = None,
    ):
        if not untransformed_features:
            untransformed_features = []
        if not transformed_features:
            transformed_features = []

        for untransformed_feature, transformed_feature in itertools.zip_longest(
            untransformed_features, transformed_features
        ):
            try:
                self._async_worker_thread.submit_task(
                    (untransformed_feature, transformed_feature)
                )
            except asyncio.QueueFull:
                _logger.error("Queue is full. Failed to log features.")

    async def _send_events(self, task):
        try:
            untransformed_feature = task[0]
            transformed_feature = task[1]

            events = []
            ce_id = str(uuid.uuid4())
            ce_time = datetime.now(timezone.utc).isoformat()
            for transformed, feature_vector in [
                (False, untransformed_feature),
                (True, transformed_feature),
            ]:
                if feature_vector:
                    events.append(
                        self._create_cloud_event(
                            ce_id,
                            ce_time,
                            self._feature_view.feature_logging.get_feature_group(
                                transformed
                            ),
                            self._avro_encode_features(
                                self._feature_encoders[transformed][0],
                                self._feature_encoders[transformed][1],
                                feature_vector,
                            ),
                        )
                    )

            responses = await get_feature_logging_client().post(
                json.dumps(events, cls=EventEncoder),
                headers=self._create_cloud_headers(),
            )
            _logger.debug(f"Feature logging events sent successfully: {responses}")
        except Exception as e:
            _logger.error(f"Failed to send events: {e}")
            traceback.print_exc()

    def _create_cloud_headers(self):
        return {
            "content-type": "application/cloudevents-batch+json; charset=UTF-8",
            "ce-specversion": "1.0",
            "ce-type": "serving.hops.works.logging.features",
            "ce-source": self._source or "http://localhost:8099",
        }

    def _create_cloud_event(self, ce_id, ce_time, fg, features):
        return {
            "specversion": "1.0",
            "type": "serving.hops.works.logging.features",
            "source": self._source,
            "id": ce_id,
            "time": ce_time,
            "datacontenttype": "application/json",
            "endpoint": "",
            "inferenceservicename": self._deployment_name,
            "namespace": self._namespace,
            "projectId": self._project_id,
            "featureGroupId": fg.id,
            "topicName": fg._online_topic_name,
            "subjectId": fg.subject["id"],
            "data": features,
        }

    def init(self, feature_view: FeatureView) -> None:
        self._feature_view = feature_view
        self._init_kafka_resource(feature_view)
        init_client(self._feature_logger_config)

        # Start worker thread after initializing the client.
        self._async_worker_thread._initialize_workers(
            self._max_concurrent_tasks, self._send_events
        )
        self._async_worker_thread.start()

    def _init_kafka_resource(self, feature_view):
        for transformed in [True, False]:
            fg = feature_view.feature_logging.get_feature_group(transformed)
            feature_writers, writer = get_writer_function(fg)
            self._feature_encoders[transformed] = (feature_writers, writer)

    def _avro_encode_features(self, complex_feature_encoder, feature_encoder, features):
        return encode_row(complex_feature_encoder, feature_encoder, features)

    @public
    def close(self):
        """Close the async feature logger."""
        # Close the async worker thread
        self._async_worker_thread.close()



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import TYPE_CHECKING, TypeVar

import numpy
from hopsworks_apigen import public
from hsml.utils.schema.columnar_schema import ColumnarSchema
from hsml.utils.schema.tensor_schema import TensorSchema


if TYPE_CHECKING:
    import pandas


@public
class Schema:
    """Create a schema for a model input or output.

    Parameters:
        object: The object to construct the schema from.
    """

    def __init__(
        self,
        object: pandas.DataFrame
        | pandas.Series
        | TypeVar("pyspark.sql.dataframe.DataFrame")
        | TypeVar("hsfs.training_dataset.TrainingDataset")
        | numpy.ndarray
        | list
        | None = None,
        **kwargs,
    ):
        # A tensor schema is either ndarray of a list containing name, type and shape dicts
        if isinstance(object, numpy.ndarray) or (
            isinstance(object, list) and all("shape" in entry for entry in object)
        ):
            self.tensor_schema = self._convert_tensor_to_schema(object).tensors
        else:
            self.columnar_schema = self._convert_columnar_to_schema(object).columns

    def _convert_columnar_to_schema(self, object):
        return ColumnarSchema(object)

    def _convert_tensor_to_schema(self, object):
        return TensorSchema(object)

    def _get_type(self):
        if hasattr(self, "tensor_schema"):
            return "tensor"
        if hasattr(self, "columnar_schema"):
            return "columnar"
        return None

    def json(self):
        return json.dumps(
            self, default=lambda o: getattr(o, "__dict__", o), sort_keys=True, indent=2
        )

    @public
    def to_dict(self):
        """Get dict representation of the Schema."""
        return json.loads(self.json())

    def __repr__(self):
        return f"Schema(type: {self._get_type()!r})"



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import BaseModel


if TYPE_CHECKING:
    from hopsworks_common.job import Job as BaseJob


class Execution(BaseModel):
    id: int
    state: str | None = None
    final_status: str | None = None
    submission_time: str | None = None
    stdout_path: str | None = None
    stderr_path: str | None = None
    app_id: str | None = None
    hdfs_user: str | None = None
    args: list[str] | None = None
    progress: float | None = None
    user: str | None = None
    files_to_remove: list[str] | None = None
    duration: float | None = None
    flink_master_url: str | None = None
    monitoring: dict | None = None
    type: str | None = None


class Job(BaseModel):
    """Model representing a job in Hopsworks MCP."""

    id: int
    name: str
    creation_time: str | None = None
    config: int
    job_type: str
    creator: str
    executions: list[Execution] | None = None
    type: str | None = None
    job_schedule: str | None = None


class Jobs(BaseModel):
    """Model representing a collection of jobs in Hopsworks MCP."""

    jobs: list[Job]
    total: int


def to_base_model_job(job: BaseJob) -> Job:
    """Convert a Job instance to a BaseModel.

    Parameters:
        job: The Job instance to convert.

    Returns:
        A Job instance as a BaseModel.
    """
    return Job(
        id=job["id"],
        name=job["name"],
        creation_time=job["creation_time"],
        config=job["config"],
        job_type=job["job_type"],
        creator=job["creator"],
        executions=[
            Execution(
                id=exec["id"],
                state=exec["state"],
                final_status=exec["final_status"],
                submission_time=exec.get("submission_time", None),
                stderr_path=exec.get("stderr_path", None),
                stdout_path=exec.get("stdout_path", None),
                app_id=exec.get("app_id", None),
                hdfs_user=exec.get("hdfs_user", None),
                args=exec.get("args", None),
                progress=exec.get("progress", None),
                user=exec.get("user", None),
                files_to_remove=exec.get("files_to_remove", None),
                duration=exec.get("duration", None),
                flink_master_url=exec.get("flink_master_url", None),
                monitoring=exec.get("monitoring", None),
                type=exec.get("type", None),
            )
            for exec in job["executions"]
        ],
        type=job["type"],
        job_schedule=job["job_schedule"],
    )



#
#   Copyright 2025 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import asyncio
import os
from pathlib import Path
from typing import TYPE_CHECKING

from fastmcp import (
    Context,  # noqa: TC002
)
from fastmcp.server.dependencies import get_context
from fastmcp.server.http import _current_http_request
from filelock import AsyncFileLock
from hopsworks.mcp.utils.tags import TAGS
from pydantic import BaseModel


if TYPE_CHECKING:
    from fastmcp import (
        FastMCP,
    )


class ExecutionResult(BaseModel):
    output: str = ""
    returncode: int | None = None


class BrewerTools:
    def __init__(self, mcp: FastMCP):
        """Initialize the BrewerTools with the MCP server instance.

        Parameters:
            mcp: The MCP server instance
        """
        self.mcp = mcp
        self.mcp.tool(tags=[TAGS.BREWER])(self.execute)

    # TODO: Use on_notification Middleware to handle cancellation requests, add process manager

    async def execute(  # docsig: disable
        self,
        chat_id: int,
        script_path: Path,
        ctx: Context,
    ) -> ExecutionResult:
        """Execute a Python script in a conda environment for a specific chat."""
        await ctx.info("Locking the chat environment for execution...\n")
        chatdir = Path(f"/hopsfs/Brewer/{chat_id}")
        async with AsyncFileLock(f"{chatdir}/.lock"):
            await extract_hopsworks_credentials(chatdir)

            envname = await get_chat_env(chat_id, chatdir)

            await ctx.info("Executing the script...\n")
            if not script_path.is_absolute():
                script_path = (chatdir / script_path).resolve()

            envcopy = os.environ.copy()
            envcopy["SECRETS_DIR"] = str(chatdir)

            proc = await asyncio.create_subprocess_exec(
                "conda",
                "run",
                "-n",
                envname,
                "python",
                script_path.name,
                cwd=script_path.parent,
                env=envcopy,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT,
            )
            if proc.stdout is None:
                raise Exception("Unable to create env: stdout is None")
            output = ""
            while proc.returncode is None:
                buf = await proc.stdout.readline()
                if not buf:
                    break
                b = buf.decode()
                output += b
                await ctx.info(b)

            return ExecutionResult(
                output=output,
                returncode=proc.returncode,
            )


async def extract_hopsworks_credentials(chatdir: Path):
    ctx = get_context()
    await ctx.info("Setting up authentication credentials...\n")
    request = _current_http_request.get()
    if not request:
        raise Exception("No HTTP request found")
    auth = next(
        (
            request.headers.get(key)
            for key in request.headers
            if key.lower() == "authorization"
        ),
        None,
    )
    if not auth:
        raise Exception("No authentication header found")
    if auth.startswith("Bearer"):
        with open(chatdir / "token.jwt", "w") as f:
            f.write(auth.removeprefix("Bearer").strip())
    elif auth.startswith("ApiKey"):
        with open(chatdir / "api.key", "w") as f:
            f.write(auth.removeprefix("ApiKey").strip())
    else:
        raise Exception("Unknown auth type")


async def get_chat_env(chat_id: int, chatdir: Path) -> str:
    ctx = get_context()
    await ctx.info("Checking if the chat environment is ready to be used...\n")
    proc = await asyncio.create_subprocess_exec(
        "conda",
        "env",
        "list",
        cwd=chatdir,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.DEVNULL,
    )
    envlist, _ = await proc.communicate()
    envname = f"chat{chat_id}"
    if envname not in envlist.decode():
        await ctx.info("Creating a new chat environment...\n")
        proc = await asyncio.create_subprocess_exec(
            "conda",
            "create",
            "--clone",
            "hopsworks_environment",
            "-n",
            envname,
            cwd=chatdir,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )
        if proc.stdout is None:
            raise Exception("Unable to create env: stdout is None")
        output = ""
        while proc.returncode is None:
            buf = await proc.stdout.readline()
            if not buf:
                break
            b = buf.decode()
            output += b
            await ctx.info(b)
        if proc.returncode:
            raise Exception(f"Unable to create env ({proc.returncode}):\n{output}")
    return envname
